#include "AABB.h"

/**
     * Check for collision on the X axis
     *
     * @param otherBoundingBox The other bounding box that is colliding with the this one.
     * @param x                Position on the X axis that is colliding
     * @return Returns the corrected x position that collided.
     */
    double AABB::clipXCollide(AABB otherBoundingBox, double x) {
        // Check if the boxes are colliding on the Y axis
        if (otherBoundingBox.maxY <= minY || otherBoundingBox.minY >= maxY) {
            return x;
        }

        // Check if the boxes are colliding on the Z axis
        if (otherBoundingBox.maxZ <= minZ || otherBoundingBox.minZ >= maxZ) {
            return x;
        }

        // Check for collision if the X axis of the current box is bigger
        if (x > 0.0F && otherBoundingBox.maxX <= minX) {
            double max = minX - otherBoundingBox.maxX - epsilon;
            if (max < x) {
                x = max;
            }
        }

        // Check for collision if the X axis of the current box is smaller
        if (x < 0.0F && otherBoundingBox.minX >= maxX) {
            double max = maxX - otherBoundingBox.minX + epsilon;
            if (max > x) {
                x = max;
            }
        }

        return x;
    }

    /**
     * Check for collision on the Y axis
     *
     * @param otherBoundingBox The other bounding box that is colliding with the this one.
     * @param y                Position on the X axis that is colliding
     * @return Returns the corrected x position that collided.
     */
double AABB::clipYCollide(AABB otherBoundingBox, double y) {
    // Check if the boxes are colliding on the X axis
    if (otherBoundingBox.maxX <= minX || otherBoundingBox.minX >= maxX) {
        return y;
    }

    // Check if the boxes are colliding on the Z axis
    if (otherBoundingBox.maxZ <= minZ || otherBoundingBox.minZ >= maxZ) {
        return y;
    }

    // Check for collision if the Y axis of the current box is bigger
    if (y > 0.0F && otherBoundingBox.maxY <= minY) {
        double max = minY - otherBoundingBox.maxY - epsilon;
        if (max < y) {
            y = max;
        }
    }

    // Check for collision if the Y axis of the current box is bigger
    if (y < 0.0F && otherBoundingBox.minY >= maxY) {
        double max = maxY - otherBoundingBox.minY + epsilon;
        if (max > y) {
            y = max;
        }
    }

    return y;
}

    /**
     * Check for collision on the Y axis
     *
     * @param otherBoundingBox The other bounding box that is colliding with the this one.
     * @param z                Position on the X axis that is colliding
     * @return Returns the corrected x position that collided.
     */
double AABB::clipZCollide(AABB otherBoundingBox, double z) {
    // Check if the boxes are colliding on the X axis
    if (otherBoundingBox.maxX <= minX || otherBoundingBox.minX >= maxX) {
        return z;
    }

    // Check if the boxes are colliding on the Y axis
    if (otherBoundingBox.maxY <= minY || otherBoundingBox.minY >= maxY) {
        return z;
    }

    // Check for collision if the Z axis of the current box is bigger
    if (z > 0.0F && otherBoundingBox.maxZ <= minZ) {
        double max = minZ - otherBoundingBox.maxZ - epsilon;
        if (max < z) {
            z = max;
        }
    }

    // Check for collision if the Z axis of the current box is bigger
    if (z < 0.0F && otherBoundingBox.minZ >= maxZ) {
        double max = maxZ - otherBoundingBox.minZ + epsilon;
        if (max > z) {
            z = max;
        }
    }

    return z;
}

/**
     * Expand the bounding box. Positive and negative numbers controls which side of the box should grow.
     *
     * @param x Amount to expand the minX or maxX
     * @param y Amount to expand the minY or maxY
     * @param z Amount to expand the minZ or maxZ
     * @return The expanded bounding box
     */
    AABB AABB::expand(double x, double y, double z) {
        double minX = this->minX;
        double minY = this->minY;
        double minZ = this->minZ;
        double maxX = this->maxX;
        double maxY = this->maxY;
        double maxZ = this->maxZ;

        // Handle expanding of min/max x
        if (x < 0.) {
            minX += x;
        } else {
            maxX += x;
        }

        // Handle expanding of min/max y
        if (y < 0.) {
            minY += y;
        } else {
            maxY += y;
        }

        // Handle expanding of min/max z
        if (z < 0.) {
            minZ += z;
        } else {
            maxZ += z;
        }

        // Create new bounding box
        return AABB(minX, minY, minZ, maxX, maxY, maxZ);
    }
