#ifndef AABB_H_INCLUDED
#define AABB_H_INCLUDED

class AABB {

	private :
    double epsilon = 0.0F;

	public :
    double minX;
    double minY;
    double minZ;
    double maxX;
    double maxY;
    double maxZ;

    /**
     * Bounding box
     *
     * @param minX Minimum x side
     * @param minY Minimum y side
     * @param minZ Minimum z side
     * @param maxX Maximum x side
     * @param maxY Maximum y side
     * @param maxZ Maximum z side
     */
    AABB(double minX, double minY, double minZ, double maxX, double maxY, double maxZ) : minX(minX), minY(minY), minZ(minZ),
                                                                                        maxX(maxX), maxY(maxY), maxZ(maxZ) {}

    /**
     * Expand the bounding box. Positive and negative numbers controls which side of the box should grow.
     *
     * @param x Amount to expand the minX or maxX
     * @param y Amount to expand the minY or maxY
     * @param z Amount to expand the minZ or maxZ
     * @return The expanded bounding box
     */
    AABB expand(double x, double y, double z);

    /**
     * Expand the bounding box on both sides.
     * The center is always fixed when using grow.
     *
     * @param x
     * @param y
     * @param z
     * @return
     */
    AABB grow(double x, double y, double z) {
        return AABB(minX - x, minY - y, minZ - z, maxX + x, maxY + y, maxZ + z);
    }

    /**
     * Check for collision on the X axis
     *
     * @param otherBoundingBox The other bounding box that is colliding with the this one.
     * @param x                Position on the X axis that is colliding
     * @return Returns the corrected x position that collided.
     */
    double clipXCollide(AABB otherBoundingBox, double x);

    /**
     * Check for collision on the Y axis
     *
     * @param otherBoundingBox The other bounding box that is colliding with the this one.
     * @param y                Position on the X axis that is colliding
     * @return Returns the corrected x position that collided.
     */
    double clipYCollide(AABB otherBoundingBox, double y);

    /**
     * Check for collision on the Y axis
     *
     * @param otherBoundingBox The other bounding box that is colliding with the this one.
     * @param z                Position on the X axis that is colliding
     * @return Returns the corrected x position that collided.
     */
    double clipZCollide(AABB otherBoundingBox, double z);

    /**
     * Check if the two boxes are intersecting/overlapping
     *
     * @param otherBoundingBox The other bounding box that could intersect
     * @return The two boxes are overlapping
     */
    bool intersects(AABB otherBoundingBox) {
        // Check on X axis
        if (otherBoundingBox.maxX <= minX || otherBoundingBox.minX >= maxX) {
            return false;
        }

        // Check on Y axis
        if (otherBoundingBox.maxY <= minY || otherBoundingBox.minY >= maxY) {
            return false;
        }

        // Check on Z axis
        return (!(otherBoundingBox.maxZ <= minZ)) && (!(otherBoundingBox.minZ >= maxZ));
    }

    /**
     * Move the bounding box relative.
     *
     * @param x Relative offset x
     * @param y Relative offset y
     * @param z Relative offset z
     */
    void move(double x, double y, double z) {
        minX += x;
        minY += y;
        minZ += z;
        maxX += x;
        maxY += y;
        maxZ += z;
    }

    /**
     * Create a new bounding box with the given offset
     *
     * @param x Relative offset x
     * @param y Relative offset x
     * @param z Relative offset x
     * @return New bounding box with the given offset relative to this bounding box
     */
    AABB offset(double x, double y, double z) {
        return AABB(minX + x, minY + y, minZ + z, maxX + x, maxY + y, maxZ + z);
    }
};


#endif // AABB_H_INCLUDED
