#ifndef FRUSTUM_H_INCLUDED
#define FRUSTUM_H_INCLUDED

#include <list>
#include <cmath>

#include <GL/gl.h>

#include "AABB.h"

#define RIGHT 0
#define LEFT 1
#define BOTTOM 2
#define TOP 3
#define BACK 4
#define FRONT 5

#define A 0
#define B 1
#define C 2
#define D 3

using namespace std;

/*
 * Author: Ben Humphrey (DigiBen)
 * E-mail: digiben@gametutorials.com
 */
class Frustum {
  private :
    static Frustum frustum;

    // This holds the A B C and D values for each side of our frustum.
    double m_Frustum[6][4];

    /**
     * doubleBuffer to get ModelView matrix.
     **/
    double modl_b[16];

    /**
     * doubleBuffer to get Projection matrix.
     **/
    double proj_b[16];


    /**
     * Frustum constructor.
     */
    public :
    Frustum() {
    }

    /**
     * Calculate the frustum and return it
     *
     * @return Current frustum
     */ static Frustum getFrustum() {
        Frustum::frustum.calculateFrustum();
        return Frustum::frustum;
    }

    ///////////////////////////////// NORMALIZE PLANE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
    /////
    /////   This normalizes a plane (A side) from a given frustum.
    /////
    ///////////////////////////////// NORMALIZE PLANE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
    void normalizePlane(double frustum[6][4], int side) {
        // Here we calculate the magnitude of the normal to the plane (point A B C)
        // Remember that (A, B, C) is that same thing as the normal's (X, Y, Z).
        // To calculate magnitude you use the equation:  magnitude = sqrt( x^2 + y^2 + z^2)
        double magnitude = (double) sqrt(frustum[side][A] * frustum[side][A] +
                frustum[side][B] * frustum[side][B] + frustum[side][C] * frustum[side][C]);

        // Then we divide the plane's values by it's magnitude.
        // This makes it easier to work with.
        frustum[side][A] /= magnitude;
        frustum[side][B] /= magnitude;
        frustum[side][C] /= magnitude;
        frustum[side][D] /= magnitude;
    }

    void calculateFrustum();

    // The code below will allow us to make checks within the frustum.  For example,
    // if we want to see if a point, a sphere, or a cube lies inside of the frustum.
    // Because all of our planes point INWARDS (The normals are all pointing inside the frustum)
    // we then can assume that if a point is in FRONT of all of the planes, it's inside.

    ///////////////////////////////// POINT IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
    /////
    /////   This determines if a point is inside of the frustum
    /////
    ///////////////////////////////// POINT IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
    bool pointInFrustum(double x, double y, double z) {
        // Go through all the sides of the frustum
        for (int i = 0; i < 6; i++) {
            // Calculate the plane equation and check if the point is behind a side of the frustum
            if (m_Frustum[i][A] * x + m_Frustum[i][B] * y + m_Frustum[i][C] * z + m_Frustum[i][D] <= 0) {
                // The point was behind a side, so it ISN'T in the frustum
                return false;
            }
        }

        // The point was inside of the frustum (In front of ALL the sides of the frustum)
        return true;
    }


    ///////////////////////////////// SPHERE IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
    /////
    /////   This determines if a sphere is inside of our frustum by it's center and radius.
    /////
    ///////////////////////////////// SPHERE IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
    bool sphereInFrustum(double x, double y, double z, double radius) {
        // Go through all the sides of the frustum
        for (int i = 0; i < 6; i++) {
            // If the center of the sphere is farther away from the plane than the radius
            if (m_Frustum[i][A] * x + m_Frustum[i][B] * y + m_Frustum[i][C] * z + m_Frustum[i][D] <= -radius) {
                // The distance was greater than the radius so the sphere is outside of the frustum
                return false;
            }
        }

        // The sphere was inside of the frustum!
        return true;
    }

    bool cubeInFrustum(double minX, double minY, double minZ, double maxX, double maxY, double maxZ);

    bool cubeInFrustum(AABB aabb) {
        return cubeInFrustum(aabb.minX, aabb.minY, aabb.minZ,
                aabb.maxX, aabb.maxY, aabb.maxZ);
    }
};

#endif // FRUSTUM_H_INCLUDED
