#ifndef LEVEL_H_INCLUDED
#define LEVEL_H_INCLUDED

#include <iostream>
#include <fstream>

#include <cmath>

#include <list>

#include "LevelListener.h"
#include "AABB.h"

using namespace std;

class Level {

  public :
    int width;
    int height;
    int depth;

  private :
    char *blocks;
    int *lightDepths;

    list<LevelListener*> levelListeners;

    Level& operator=(const Level &a) {(void)a; return *this;}

	public :

	/**
     * Copy constructor
     *
     * @param a Other Level
     */
	Level(Level &a) : width(a.width), height(a.height), depth(a.depth), blocks(new char[a.width * a.height * a.depth]),
                    lightDepths(new int[a.width * a.height]), levelListeners(a.levelListeners) {
        for(int i=0; i < width * height * depth; i++)
            blocks[i] = a.blocks[i];
        for(int i=0; i < width * depth; i++)
            lightDepths[i] = a.lightDepths[i];
    }

    /**
     * Three dimensional level containing all tiles
     *
     * @param width  Level width
     * @param height Level height
     * @param depth  Level depth
     */
    Level(int width, int height, int depth) : width(width), height(height), depth(depth),
                                            blocks(new char[width * height * depth]), lightDepths(new int[width * height]),
                                            levelListeners() {

        // Fill level with tiles
        for (int x = 0; x < width; x++) {
            for (int y = 0; y < depth; y++) {
                for (int z = 0; z < height; z++) {
                    // Calculate index from x, y and z
                    int index = (y * this->height + z) * this->width + x;

                    // Fill level with tiles
                    blocks[index] = (char) ((y <= depth * 2 / 3) ? 1 : 0);
                }
            }
        }

        // Calculate light depth of the entire level
        calcLightDepths(0, 0, width, height);

        // Load level if it exists
        load();
    }

    /**
     * Load blocks from level.dat
     */
    void load() {
		    ifstream dis("./level.mcp", ios::in | ios::binary);
		    dis.read(blocks, width * height * depth);
		    calcLightDepths(0, 0, width, height);
		    dis.close();

		    // Notify all tiles changed
		    for (list<LevelListener*>::iterator i = levelListeners.begin(); i != levelListeners.end(); i++) {
		        (**i).allChanged();
		    }
    }

    /**
     * Store blocks in level.dat
     */
    void save() {
		    ofstream dos("level.dat", ios::out | ios::binary);
		    dos.write(blocks, width * height * depth);
		    dos.close();
    }

	private :
    void calcLightDepths(int minX, int minZ, int maxX, int maxZ);

	public :
    /**
     * Return true if a tile is available at the given location
     *
     * @param x Level position x
     * @param y Level position y
     * @param z Level position z
     * @return Tile available
     */
    bool isTile(int x, int y, int z) {
        // Is location out of the level?
        if (x < 0 || y < 0 || z < 0 || x >= width || y >= depth || z >= height) {
            return false;
        }

        // Calculate index from x, y and z
        int index = (y * height + z) * width + x;

        // Return true if there is a tile at this location
        return blocks[index] != 0;
    }

    /**
     * Returns true if tile is solid and not transparent
     *
     * @param x Tile position x
     * @param y Tile position y
     * @param z Tile position z
     * @return Tile is solid
     */
    bool isSolidTile(int x, int y, int z) {
        return isTile(x, y, z);
    }

    /**
     * Returns true if the tile is blocking the light
     *
     * @param x Tile position x
     * @param y Tile position y
     * @param z Tile position z
     * @return Tile blocks the light
     */
    bool isLightBlocker(int x, int y, int z) {
        return isSolidTile(x, y, z);
    }

    /**
     * Get brightness of a tile
     *
     * @param x Tile position x
     * @param y Tile position y
     * @param z Tile position z
     * @return The brightness value from 0 to 1
     */
    float getBrightness(int x, int y, int z) {
        // Define brightness
        float dark = .8;
        float light = 1.;

        // Is light tile
        if (x < 0 || y < 0 || z < 0 || x >= width || y >= depth || z >= height) {
            return light;
        }

        // Is dark tile
        if (y < lightDepths[x + z * width]) {
            return dark;
        }

        // Unknown brightness
        return light;
    }


    list<AABB> getCubes(AABB boundingBox);


    /**
     * Set tile at position
     *
     * @param x  Tile position x
     * @param y  Tile position y
     * @param z  Tile position z
     * @param id Type of tile
     */
    void setTile(int x, int y, int z, int id) {
        // Check if position is out of level
        if (x < 0 || y < 0 || z < 0 || x >= width || y >= depth || z >= height) {
            return;
        }

        // Set tile
        blocks[(y * height + z) * width + x] = (char) id;

        // Update lightning
        calcLightDepths(x, z, 1, 1);

        // Notify tile changed
        for (list<LevelListener*>::iterator i = levelListeners.begin(); i != levelListeners.end(); i++) {
            (**i).tileChanged(x, y, z);
        }
    }

    /**
     * Register a level listener
     *
     * @param levelListener Listener interface
     */
    void addListener(LevelListener *levelListener) {
        levelListeners.push_back(levelListener);
    }
};

#endif // LEVEL_H_INCLUDED
