#include "LevelRenderer.h"

    /**
     * Create renderer for level
     *
     * @param level The rendered level
     */
LevelRenderer::LevelRenderer(Level *level) : tessellator(), level(level), chunks(nullptr),
                                            chunkAmountX(level->width / CHUNK_SIZE), chunkAmountY(level->depth / CHUNK_SIZE),
                                            chunkAmountZ(level->height / CHUNK_SIZE){
    chunks = new Chunk *[chunkAmountX * chunkAmountY * chunkAmountZ];
    level->addListener(this);

    // Fill level with chunks
    for (int x = 0; x < chunkAmountX; x++) {
        for (int y = 0; y < chunkAmountY; y++) {
            for (int z = 0; z < chunkAmountZ; z++) {
                // Calculate min bounds for chunk
                int minChunkX = x * CHUNK_SIZE;
                int minChunkY = y * CHUNK_SIZE;
                int minChunkZ = z * CHUNK_SIZE;

                // Calculate max bounds for chunk
                int maxChunkX = (x + 1) * CHUNK_SIZE;
                int maxChunkY = (y + 1) * CHUNK_SIZE;
                int maxChunkZ = (z + 1) * CHUNK_SIZE;

                // Check for chunk bounds out of level
                maxChunkX = min(level->width, maxChunkX);
                maxChunkY = min(level->depth, maxChunkY);
                maxChunkZ = min(level->height, maxChunkZ);

                // Create chunk based on bounds
                Chunk *chunk = new Chunk(*level, minChunkX, minChunkY, minChunkZ, maxChunkX, maxChunkY, maxChunkZ);
                chunks[(x + y * chunkAmountX) * chunkAmountZ + z] = chunk;
            }
        }
    }
}

    /**
     * Mark all chunks inside of the given area as dirty.
     *
     * @param minX Minimum on X axis
     * @param minY Minimum on Y axis
     * @param minZ Minimum on Z axis
     * @param maxX Maximum on X axis
     * @param maxY Maximum on Y axis
     * @param maxZ Maximum on Z axis
     */
void LevelRenderer::setDirty(int minX, int minY, int minZ, int maxX, int maxY, int maxZ) {
    // To chunk coordinates
    minX /= CHUNK_SIZE;
    minY /= CHUNK_SIZE;
    minZ /= CHUNK_SIZE;
    maxX /= CHUNK_SIZE;
    maxY /= CHUNK_SIZE;
    maxZ /= CHUNK_SIZE;

    // Minimum limit
    minX = max(minX, 0);
    minY = max(minY, 0);
    minZ = max(minZ, 0);

    // Maximum limit
    maxX = min(maxX, chunkAmountX - 1);
    maxY = min(maxY, chunkAmountY - 1);
    maxZ = min(maxZ, chunkAmountZ - 1);

    // Mark all chunks as dirty
    for (int x = minX; x <= maxX; x++) {
        for (int y = minY; y <= maxY; y++) {
            for (int z = minZ; z <= maxZ; z++) {
                // Set dirty
                chunks[(x + y * chunkAmountX) * chunkAmountZ + z]->setDirty();
            }
        }
    }
}

    /**
     * Render pick selection face on tile
     *
     * @param player The player
     */
void LevelRenderer::pick(Player &player) {
    float radius = 3.;
    AABB boundingBox = player.boundingBox.grow(radius, radius, radius);

    int minX = (int) boundingBox.minX;
    int maxX = (int) (boundingBox.maxX + 1.);
    int minY = (int) boundingBox.minY;
    int maxY = (int) (boundingBox.maxY + 1.);
    int minZ = (int) boundingBox.minZ;
    int maxZ = (int) (boundingBox.maxZ + 1.);

    glInitNames();
    for (int x = minX; x < maxX; x++) {
        // Name value x
        glPushName(x);
        for (int y = minY; y < maxY; y++) {
            // Name value y
            glPushName(y);
            for (int z = minZ; z < maxZ; z++) {
                // Name value z
                glPushName(z);
                // Check for solid tile
                if (level->isSolidTile(x, y, z)) {
                    // Name value type
                    glPushName(0);
                    // Render all faces
                    for (int face = 0; face < 6; face++) {
                        // Name value face id
                        glPushName(face);
                        // Render selection face
                        tessellator.init();
                        Tile::rock.renderFace(tessellator, x, y, z, face);
                        tessellator.flush();
                        glPopName();
                    }
                    glPopName();
                }
                glPopName();
            }
            glPopName();
        }
        glPopName();
    }
}

const int LevelRenderer::CHUNK_SIZE = 16;
