#ifndef PLAYER_H_INCLUDED
#define PLAYER_H_INCLUDED

#include "Keyboard.h"
#include "Level.h"
#include "AABB.h"

#include <cstdlib>
#include <list>
#include <cmath>

using namespace std;

class Player {

    Level *level;
    Keyboard keyboard;

    Player(Player &a) : level(nullptr), keyboard(), x(0.), y(0.), z(0.), prevX(0.), prevY(0.),
                        prevZ(0.), motionX(0.), motionY(0.), motionZ(0.), xRotation(0.), yRotation(0.),
                        onGround(false), boundingBox(0., 0., 0., 0., 0., 0.) {(void)a;}

    Player &operator=(const Player &a){(void) a; return *this;}

  public :
    double x, y, z;
    double prevX, prevY, prevZ;
    double motionX, motionY, motionZ;
    double xRotation, yRotation;

    bool onGround;

    AABB boundingBox;

    /**
     * The player that is controlling the camera of the game
     *
     * @param level Level of the player
     * @param keyboard Logical Keyboard used to make things easier
     */
    Player(Level *level, Keyboard keyboard) : level(level), keyboard(keyboard), x(0.), y(0.), z(0.), prevX(0.), prevY(0.),
                                            prevZ(0.), motionX(0.), motionY(0.), motionZ(0.), xRotation(0.), yRotation(0.),
                                            onGround(false), boundingBox(0., 0., 0., 0., 0., 0.) {
        resetPosition();
    }

    /**
     * Set the player to a specific location
     *
     * @param x Position x
     * @param y Position y
     * @param z Position z
     */
   private :
    void setPosition(double x, double y, double z) {
        // Set the position of the player
        this->x = x;
        this->y = y;
        this->z = z;

        // Player size
        double width = .3;
        double height = .9;

        // Set the position of the bounding box
        boundingBox = AABB(x - width, y - height,
                z - width, x + width,
                y + height, z + width);
    }

    /**
     * Reset the position of the player to a random location on the level
     */
    void resetPosition() {
        double x = (double) rand() * level->width / RAND_MAX;
        double y = (double) (level->depth + 3);
        double z = (double) rand() * level->height / RAND_MAX;

        setPosition(x, y, z);
    }

    /**
     * Turn the camera using motion yaw and pitch
     *
     * @param x Rotate the camera using yaw
     * @param y Rotate the camera using pitch
     */
  public :
    void turn(double x, double y) {
        yRotation += x * .15;
        xRotation -= y * .15;

        // Pitch limit
        xRotation = max(-90., xRotation);
        xRotation = min(90., xRotation);
    }

    void tick();

    void move(double x, double y, double z);

    /**
     * Add motion to the player in the facing direction with given speed
     *
     * @param x     Motion to add on X axis
     * @param z     Motion to add on Z axis
     * @param speed Strength of the added motion
     */
  private :
    void moveRelative(double x, double z, double speed) {
        double distance = x * x + z * z;

        // Stop moving if too slow
        if (distance < .01)
            return;

        // Apply speed to relative movement
        distance = speed / sqrt(distance);
        x *= distance;
        z *= distance;

        // Calculate sin and cos of player rotation
        double _sin = sin(yRotation*3.14159265/180);
        double _cos = cos(yRotation*3.14159265/180);

        // Move the player in facing direction
        motionX += x * _cos - z * _sin;
        motionZ += z * _cos + x * _sin;
    }
};


#endif // PLAYER_H_INCLUDED
