#include "Tile.h"

Tile Tile::grass = Tile(0);
Tile Tile::rock = Tile(1);

    /**
     * Render a tile at the given position
     *
     * @param tessellator Tessellator for rendering
     * @param level       Level to check for surrounding tiles
     * @param layer       The layer which decides if it's a shadow or not
     * @param x           Tile position x
     * @param y           Tile position y
     * @param z           Tile position z
     */
void Tile::render(Tessellator tessellator, Level &level, int layer, int x, int y, int z) {
    float minU = 0.;
    float maxU = 1.;
    float minV = 0.;
    float maxV = 1.;

    float shadeX = .6;
    float shadeY = 1.;
    float shadeZ = .8;

    float minX = x + 0.;
    float maxX = x + 1.;
    float minY = y + 0.;
    float maxY = y + 1.;
    float minZ = z + 0.;
    float maxZ = z + 1.;

    // Render bottom face
    if (!level.isSolidTile(x, y - 1, z)) {
        // Get the brightness of the tile below
        float brightness = level.getBrightness(x, y - 1, z) * shadeY;

        // Don't render face if both conditions are the same (isShadowLayer != isFullBrightness)
        if ((layer == 1) ^ (brightness == shadeY)) {
            tessellator.color(brightness, brightness, brightness);
            tessellator.texture(minU, maxV);
            tessellator.vertex(minX, minY, maxZ);
            tessellator.texture(minU, minV);
            tessellator.vertex(minX, minY, minZ);
            tessellator.texture(maxU, minV);
            tessellator.vertex(maxX, minY, minZ);
            tessellator.texture(maxU, maxV);
            tessellator.vertex(maxX, minY, maxZ);
        }
    }

    // Render top face
    if (!level.isSolidTile(x, y + 1, z)) {
        // Get the brightness of the tile above
        float brightness = level.getBrightness(x, y + 1, z) * shadeY;

        // Don't render face if both conditions are the same (isShadowLayer != isFullBrightness)
        if ((layer == 1) ^ (brightness == shadeY)) {
            tessellator.color(brightness, brightness, brightness);
            tessellator.texture(maxU, maxV);
            tessellator.vertex(maxX, maxY, maxZ);
            tessellator.texture(maxU, minV);
            tessellator.vertex(maxX, maxY, minZ);
            tessellator.texture(minU, minV);
            tessellator.vertex(minX, maxY, minZ);
            tessellator.texture(minU, maxV);
            tessellator.vertex(minX, maxY, maxZ);
        }
    }

    // Render side faces Z
    if (!level.isSolidTile(x, y, z - 1)) {
        // Get the brightness of the tile next to it
        float brightness = level.getBrightness(x, y, z - 1) * shadeZ;

        // Don't render face if both conditions are the same (isShadowLayer != isFullBrightness)
        if ((layer == 1) ^ (brightness == shadeZ)) {
            tessellator.color(brightness, brightness, brightness);
            tessellator.texture(maxU, minV);
            tessellator.vertex(minX, maxY, minZ);
            tessellator.texture(minU, minV);
            tessellator.vertex(maxX, maxY, minZ);
            tessellator.texture(minU, maxV);
            tessellator.vertex(maxX, minY, minZ);
            tessellator.texture(maxU, maxV);
            tessellator.vertex(minX, minY, minZ);
        }
    }
    if (!level.isSolidTile(x, y, z + 1)) {
        // Get the brightness of the tile next to it
        float brightness = level.getBrightness(x, y, z + 1) * shadeZ;

        // Don't render face if both conditions are the same (isShadowLayer != isFullBrightness)
        if ((layer == 1) ^ (brightness == shadeZ)) {
            tessellator.color(brightness, brightness, brightness);
            tessellator.texture(minU, minV);
            tessellator.vertex(minX, maxY, maxZ);
            tessellator.texture(minU, maxV);
            tessellator.vertex(minX, minY, maxZ);
            tessellator.texture(maxU, maxV);
            tessellator.vertex(maxX, minY, maxZ);
            tessellator.texture(maxU, minV);
            tessellator.vertex(maxX, maxY, maxZ);
        }
    }

    // Render side faces X
    if (!level.isSolidTile(x - 1, y, z)) {
        // Get the brightness of the tile next to it
        float brightness = level.getBrightness(x - 1, y, z) * shadeX;

        // Don't render face if both conditions are the same (isShadowLayer != isFullBrightness)
        if ((layer == 1) ^ (brightness == shadeX)) {
            tessellator.color(brightness, brightness, brightness);
            tessellator.texture(maxU, minV);
            tessellator.vertex(minX, maxY, maxZ);
            tessellator.texture(minU, minV);
            tessellator.vertex(minX, maxY, minZ);
            tessellator.texture(minU, maxV);
            tessellator.vertex(minX, minY, minZ);
            tessellator.texture(maxU, maxV);
            tessellator.vertex(minX, minY, maxZ);
        }
    }
    if (!level.isSolidTile(x + 1, y, z)) {
        // Get the brightness of the tile next to it
        float brightness = level.getBrightness(x + 1, y, z) * shadeX;

        // Don't render face if both conditions are the same (isShadowLayer != isFullBrightness)
        if ((layer == 1) ^ (brightness == shadeX)) {
            tessellator.color(brightness, brightness, brightness);
            tessellator.texture(minU, maxV);
            tessellator.vertex(maxX, minY, maxZ);
            tessellator.texture(maxU, maxV);
            tessellator.vertex(maxX, minY, minZ);
            tessellator.texture(maxU, minV);
            tessellator.vertex(maxX, maxY, minZ);
            tessellator.texture(minU, minV);
            tessellator.vertex(maxX, maxY, maxZ);
        }
    }
}

    /**
     * Render the single face of a tile
     *
     * @param tessellator Tessellator for rendering
     * @param x Tile position x
     * @param y Tile position y
     * @param z Tile position z
     * @param face Face id (0:Top, 1:Bottom, ...)
     */
void Tile::renderFace(Tessellator tessellator, int x, int y, int z, int face) {
    float minX = x + 0.;
    float maxX = x + 1.;
    float minY = y + 0.;
    float maxY = y + 1.;
    float minZ = z + 0.;
    float maxZ = z + 1.;

    // Render face
    if (face == 0) {
        tessellator.vertex(minX, minY, maxZ);
        tessellator.vertex(minX, minY, minZ);
        tessellator.vertex(maxX, minY, minZ);
        tessellator.vertex(maxX, minY, maxZ);
    }
    if (face == 1) {
        tessellator.vertex(maxX, maxY, maxZ);
        tessellator.vertex(maxX, maxY, minZ);
        tessellator.vertex(minX, maxY, minZ);
        tessellator.vertex(minX, maxY, maxZ);
    }
    if (face == 2) {
        tessellator.vertex(minX, maxY, minZ);
        tessellator.vertex(maxX, maxY, minZ);
        tessellator.vertex(maxX, minY, minZ);
        tessellator.vertex(minX, minY, minZ);
    }
    if (face == 3) {
        tessellator.vertex(minX, maxY, maxZ);
        tessellator.vertex(minX, minY, maxZ);
        tessellator.vertex(maxX, minY, maxZ);
        tessellator.vertex(maxX, maxY, maxZ);
    }
    if (face == 4) {
        tessellator.vertex(minX, maxY, maxZ);
        tessellator.vertex(minX, maxY, minZ);
        tessellator.vertex(minX, minY, minZ);
        tessellator.vertex(minX, minY, maxZ);
    }
    if (face == 5) {
        tessellator.vertex(maxX, minY, maxZ);
        tessellator.vertex(maxX, minY, minZ);
        tessellator.vertex(maxX, maxY, minZ);
        tessellator.vertex(maxX, maxY, maxZ);
    }
}
