/* A simple program to show how to set up an X window for OpenGL rendering.
 * X86 compilation: gcc -o -L/usr/X11/lib   main main.c -lGL -lX11
 * X64 compilation: gcc -o -L/usr/X11/lib64 main main.c -lGL -lX11
 */
#include <stdio.h>
#include <stdlib.h>
#include <exception>

#include <GL/glx.h>    /* this includes the necessary X headers */
#include <GL/gl.h>
#include <GL/glu.h>

#include <X11/X.h>    /* X11 constant (e.g. TrueColor) */
#include <X11/keysym.h>

#include "Keyboard.h"
#include "Timer.h"
#include "Level.h"
#include "LevelRenderer.h"
#include "Player.h"
#include "HitResult.h"

static int snglBuf[] = {GLX_RGBA, GLX_CONTEXT_MAJOR_VERSION_ARB, 2, GLX_CONTEXT_MINOR_VERSION_ARB, 1, GLX_DEPTH_SIZE, 16, None};
static int dblBuf[]  = {GLX_RGBA, GLX_CONTEXT_MAJOR_VERSION_ARB, 2, GLX_CONTEXT_MINOR_VERSION_ARB, 1, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER, None};

int width = 1024;
int height = 768;

Display   *dpy;
Window     win;
GLfloat    xAngle = 42.0, yAngle = 82.0, zAngle = 112.0;
GLboolean  doubleBuffer = GL_TRUE;

bool doExit = false;

Timer timer(60);
Level *level;
LevelRenderer *levelRenderer;
Player *player;
Keyboard *kbrd;
HitResult *hitResult;
int ViewportBuffer[16];
unsigned selectBuffer[2000];
float fogColor [4] = {	14 / 255.f,
                				11 / 255.f,
                				10 / 255.f,
                				1.f};

    /**
     * Game tick, called exactly 20 times per second
     */
void tick() {
    player->tick();
}

    /**
     * Move and rotate the camera to players location and rotation
     *
     * @param partialTicks Overflow ticks to calculate smooth a movement
     */
void moveCameraToPlayer(float partialTicks) {
    // Eye height
    glTranslatef(0.0f, 0.0f, -0.3f);

    // Rotate camera
    glRotatef(player->xRotation, 1.0f, 0.0f, 0.0f);
    glRotatef(player->yRotation, 0.0f, 1.0f, 0.0f);

    // Smooth movement
    double x = player->prevX + (player->x - player->prevX) * partialTicks;
    double y = player->prevY + (player->y - player->prevY) * partialTicks;
    double z = player->prevZ + (player->z - player->prevZ) * partialTicks;

    // Move camera to players location
    glTranslated(-x, -y, -z);
}


    /**
     * Setup the normal player camera
     *
     * @param partialTicks Overflow ticks to calculate smooth a movement
     */
void setupCamera(float partialTicks) {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // Set camera perspective
    gluPerspective(70, width / (float) height, 0.05F, 1000);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Move camera to middle of level
    moveCameraToPlayer(partialTicks);
}

    /**
     * Setup tile picking camera
     *
     * @param partialTicks Overflow ticks to calculate smooth a movement
     * @param x            Screen position x
     * @param y            Screen position y
     */
void setupPickCamera(float partialTicks, int x, int y) {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // Get viewport value
    glGetIntegerv(GL_VIEWPORT, ViewportBuffer);

//    // Flip
//    this.viewportBuffer.flip();
//    this.viewportBuffer.limit(16);

    // Set matrix and camera perspective
    gluPickMatrix(x, y, 5.0f, 5.0f, ViewportBuffer);
    gluPerspective(70.0f, width / (float) height, 0.05f, 1000.0f);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Move camera to middle of level
    moveCameraToPlayer(partialTicks);
}

    /**
     * @param partialTicks Overflow ticks to calculate smooth a movement
     */
void pick(float partialTicks) {
    glSelectBuffer(sizeof(selectBuffer)/sizeof(*selectBuffer),selectBuffer);
    glRenderMode(GL_SELECT);

    // Setup pick camera
    setupPickCamera(partialTicks, width / 2, height / 2);

    // Render all possible pick selection faces to the target
    levelRenderer->pick(*player);

    //// Flip buffer
    //selectBuffer.flip();

    long closest = 0L;
    int *names = new int[10];
    int hitNameCount = 0;

    // Get amount of hits
    int hits = glRenderMode(GL_RENDER);
    for (int hitIndex = 0; hitIndex < hits; hitIndex++) {

        // Get name count
        int nameCount = selectBuffer[0];
        long minZ = selectBuffer[1];

        // Check if the hit is closer to the camera
        if (minZ < closest || hitIndex == 0) {
            closest = minZ;
            hitNameCount = nameCount;

            // Fill names
            for (int nameIndex = 0; nameIndex < nameCount; nameIndex++) {
                names[nameIndex] = selectBuffer[nameIndex+2];
            }
        }/* else { // Not useful, it's not an IntBuffer :)
            // Skip names
            for (int nameIndex = 0; nameIndex < nameCount; ++nameIndex) {
                this.selectBuffer.get();
            }
        }*/
    }

    // Update hit result
    if (hitNameCount > 0) {
        hitResult = new HitResult(names[0], names[1], names[2], names[3], names[4]);
    } else {
        hitResult = nullptr;
    }
}


void render(float partialTicks, Display *dpy) {
        // Get mouse motion
        static float X;
        static float Y;
        static XEvent event;;

        // Rotate the camera using the mouse motion input
        player->turn(X, Y);

        // Pick tile
        pick(partialTicks);

        do
				{
				  XNextEvent(dpy, &event);
				  //printf("New event !\n");
				  GLenum err;
				  while((err = glGetError()) != GL_NO_ERROR)
                    printf("Current error : %d\n",err);
				  switch (event.type)
				  {
				  	case MotionNotify:
				  	{
				  		X = ((XMotionEvent *) &event)->x;
				  		Y = ((XMotionEvent *) &event)->y;
				  		XWarpPointer(dpy, None, win, 0, 0, 0, 0, width/2, height/2);
				  		break;
				  	}
				    case KeyPress:
				    case KeyRelease:
				    {
				      if (XLookupKeysym(&event.xkey, 0) == XK_Escape) {
				      		doExit = true;
				        }
				      if (XLookupKeysym(&event.xkey, 0) == XK_Return) {
				            // Save the level
				            level->save();
				        }
                      kbrd->setPressed(((XKeyEvent*)&event)->keycode,event.type==KeyPress);
				      break;
				    }
				    case ButtonPress:
				      switch (event.xbutton.button)
				      {
				        case 0 :
		              // Destroy the tile
		              if(hitResult)
                        level->setTile(hitResult->x, hitResult->y, hitResult->z, 0);
		             	break;

		            case 1 :
                // Get target tile position
                if(hitResult) {
                    int x = hitResult->x;
                    int y = hitResult->y;
                    int z = hitResult->z;

                    // Get position of the tile using face direction
                    if (hitResult->face == 0) y--;
                    if (hitResult->face == 1) y++;
                    if (hitResult->face == 2) z--;
                    if (hitResult->face == 3) z++;
                    if (hitResult->face == 4) x--;
                    if (hitResult->face == 5) x++;

                    // Set the tile
                    level->setTile(x, y, z, 1);
                }
                break;
				      }
				      break;
				    case ConfigureNotify:
				      /*glViewport(0, 0, event->xconfigure.width,
				                 event->xconfigure.height);*/
                      width = event.xconfigure.width;
                      height = event.xconfigure.height;
				      break;
				  }
				} while(XPending(dpy)); /* loop to compress events */

        // Clear color and depth buffer and reset the camera
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Setup normal player camera
        setupCamera(partialTicks);

        // Setup fog
        glEnable(GL_FOG);
        glFogi(GL_FOG_MODE, GL_LINEAR);
        glFogf(GL_FOG_START, -10);
        glFogf(GL_FOG_END, 20);
        glFogfv(GL_FOG_COLOR, fogColor);
        glDisable(GL_FOG);

        // Render bright tiles
        levelRenderer->render(0);

        // Enable fog to render shadow
        glEnable(GL_FOG);

        // Render dark tiles in shadow
        levelRenderer->render(1);

        // Finish rendering
        glDisable(GL_TEXTURE_2D);

        // Render the actual hit
        if (hitResult != nullptr) {
            levelRenderer->renderHit(*hitResult);
        }

        glDisable(GL_FOG);
        glFinish();
        glFlush();
        glXSwapBuffers(dpy, win);
    }

int main(int argc, char **argv)
{
try {
  XVisualInfo         *vi=nullptr;
  Colormap             cmap;
  XSetWindowAttributes swa;
  GLXContext           cx;
  //int                  dummy;

  /*** (1) open a connection to the X server ***/

  dpy = XOpenDisplay(NULL);

  /*** (2) find an appropriate visual ***/

  /* find an OpenGL-capable RGB visual with depth buffer */
  vi = glXChooseVisual(dpy, DefaultScreen(dpy), dblBuf);
  if (vi == NULL)
  {
    vi = glXChooseVisual(dpy, DefaultScreen(dpy), snglBuf);
    doubleBuffer = GL_FALSE;
  }

  /*** (3) create an OpenGL rendering context  ***/

  /* create an OpenGL rendering context */
  cx = glXCreateContext(dpy, vi, /* no shared dlists */ None,
                        /* direct rendering if possible */ GL_TRUE);

  /*** (4) create an X window with the selected visual ***/

  /* create an X colormap since probably not using default visual */
  cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual, AllocNone);
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = KeyPressMask    | ExposureMask | KeyReleaseMask | PointerMotionMask
                 | ButtonPressMask | StructureNotifyMask;
  win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0,
                      width, height, 0, vi->depth, InputOutput, vi->visual,
                      CWBorderPixel | CWColormap | CWEventMask, &swa);
  XSetStandardProperties(dpy, win, "main", "main", None,
                         argv, argc, NULL);

  /*** (5) bind the rendering context to the window ***/

  glXMakeCurrent(dpy, win, cx);

  /*** (6) request the X window to be displayed on the screen ***/

  XMapWindow(dpy, win);

  /*** (7) configure the OpenGL context for rendering ***/

  glEnable(GL_TEXTURE_2D);
  glShadeModel(GL_SMOOTH);
  glClearColor(0.5F, 0.8F, 1.0F, 0.0F);
  glClearDepth(1.0);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glDepthFunc(GL_LEQUAL);

  printf("GL version : %s",glGetString(GL_VERSION));

  level = new Level(256, 256, 64);
  levelRenderer = new LevelRenderer(level);
  kbrd = new Keyboard();
  player = new Player(level,*kbrd);

  /*** (8) dispatch X events ***/

  int frames = 0;

  while (!doExit)
  {
    timer.advanceTime();

    for (int i = 0; i < timer.ticks; ++i) {
        player->tick();
    }

    render(timer.partialTicks, dpy);

    frames++;
    clock_t lastTime = clock()*1000/CLOCKS_PER_SEC;

    while (clock()*1000/CLOCKS_PER_SEC >= lastTime + 1000) {
        // Print amount of frames
        printf("%d fps, %d chunk updates\n",frames, Chunk::updates);

        // Reset global rebuild stats
        Chunk::updates = 0;

        // Increase last time printed and reset frame counter
        lastTime += 1000;
        frames = 0;
    }
    //printf("%ld\n",clock());
  }
  level->save();
}
catch(exception &e) {
    printf("Exception caught !\n");
    printf("%s",e.what());
    return -1;
}

  return 0;
}
